// $Id: coverage.C,v 1.5 2000/03/21 12:38:17 geert Exp $


#include <assert.h>
#include <PLAGEO/pgn.h>
#include <PLAGEO/sweeplist.h>
#include <PLAGEO/intersect.h>
#include <PLAGEO/glwin.h>
#include <dcel.h>
#include <geolib.h>

//#define VISUALISE 1

#define debug_assert(expr) assert(expr)
//#define debug_assert(expr) 


const int UPPERBOUND = 1;
const int LOWERBOUND = -1;


typedef int vtident;	    // identifier type for vertices during the sweep.

typedef reray<pl_line> pl_linereray;

struct qnode;			    // forward declaration

typedef qnode *queueid;		    // identifier for elements in the queue.

const queueid no_event = 0;	    // value for marking absence of queue event.

inline int is_no_event(queueid qid) // test function for absence of queue event.
{ return qid==0; }


// ************* declaration of union find structure *************

/*
 * A uf_struct (or union find structure) is a structure to build and identify
 * equivalence classes.
 * `new_handle' returns a new identifier (handle) that until now was unused in
 * this union find structure. Initially, the equivalence class of this new 
 * handle contains no other handles.
 * `unite' unites two union-find handles. The equivalence classes of the two
 * handles are merged. Those handles must have been created before by a call
 * to new_handle.
 * `find' takes a handle as parameter and returns an integer: the equivalence
 * class to which this handle belongs. The handle must have been created
 * before by a new_handle call.
 * `all_ids' fills the argument reray with a number of integers. Those integers
 * are exactly all the integers that can be returned by find.
 */

typedef int uf_handle;			// identifier of union find structure.
const uf_handle INVALID_HANDLE = -1;	// value for absence of identifier.

struct uf_struct {
    uf_handle	new_handle();
    void	unite(uf_handle, uf_handle);
    int		find(uf_handle) const;
    void	all_ids(intreray &ids) const;
protected:
    mutable intreray _rr;
    bool is_valid_handle(uf_handle) const;
    bool is_equivalence_class(int i) const ;
};

/*
 * Often some information must be associated with an equivalence class.  The
 * two following classes specialise the union find structure for this purpose.
 */

struct region_uf_struct: public uf_struct {
    void	    associate(uf_handle, dcel_cell_id, uf_handle comph);
			// associate the cell identifier of a dcel with this
			// region and the surrounding component handle.
    dcel_cell_id    get_cell_id(int equivalence_class) const;
			// return the associated dcel cell identifier.
    uf_handle	    get_surrounding_comp(int equivalence_class) const;
			// return the associated surrounding component handle.
protected:
    intreray _cell_id;
    intreray _outer_comp;
};

struct component_uf_struct: public uf_struct {
    void	    associate(uf_handle, dcel_edge_id, uf_handle);
			// associate the dcel edge identifier with this
			// component and the surrounding cell handle.
    dcel_edge_id    get_outer_edge(int equivalence_class) const;
			// return the associated dcel edge. This edge lies on
			// the outside of the component.
    uf_handle	    get_surrounding_cell(int equivalence_class) const ;
			// return the associated surrounding cell handle.
protected:
    intreray _outer_edge;
    intreray _outer_cell;
};


// ************* tree structure *************

// The nodes in the tree contain a chain of edges.
// Every node stores information about this chain and about the region
// directly below it.
// Therefore there is a symbolical upper chain that is above all the other 
// chains and stores the information of the region above the normal chains.
// This symbolical chain is marked by a special value.

struct treeinfo {
// info associated with this chain.
    pl_linereray    edges;
			// the edges of a chain. Each edge is represented
			// by a line. The reference point of this line
			// is the endpoint of the edge. The direction of
			// line is the same as the direction of the edge.
    int		    orientation;
			// UPPERBOUND if the polygon lies below this chain.
			// LOWERBOUND if the polygon lies above this chain.
    int		    edgeindex;
			// Index in edges reray. Indicates the current edge.
    vtident	    end;
			// The identifier for the last vertex on this chain.
			// It has value -1 for the symbolic highest chain.
    dcel_edge_id    upedge, downedge;
			// Identifiers of current dcel edges on this chain.
			// Have value No_dcel_edge_id if non-existent.
    bool	    at_end() const;
    const pl_vec &  end_vt() const ;
			// End vertex of current edge of this chain.
// info associated with region below this chain.
    queueid	    qid;
			// crossing or delete event that ends this region.
    int		    _coverage;
			// The number of polygons that cover this region.
    uf_handle	    _region;
			// Identifying handle for this region.
    bool	    is_highest() const {return end == -1;}
};

bool treeinfo::at_end() const
{
    return edges.size() == edgeindex+1;
}

const pl_vec &treeinfo::end_vt() const
{
    return edges[edgeindex].ref_vt();
}

inline treeinfo *conv(void *p)
{   return (treeinfo *)p; }


static int tree_cmp(const pl_vec &vt, void *info)
{
    if (conv(info)->is_highest())
	return -1;
    const pl_line &line = conv(info)->edges[conv(info)->edgeindex];
    if (is_to_the_left(vt, line))   // is_above
	return 1;
    else
	return -1;
}

// ************* declaration of event queue *************


struct qnode {
friend class queue;
	    qnode(const pl_vec &vt);
    pl_vec  _vt;
protected:
    qnode   *left, *right;  // pointer to two elements later in the queue.
	// queue invariant:
	//  (left == 0 || left->_vt > _vt) && (right == 0 || right->_vt > _vt)
    int	    _deleted;	    // marks if this item is removed.
};

qnode::qnode(const pl_vec &vt)
{
    _vt = vt;
    _deleted = 0;
}

/*
 * This queue is a reasonably straightforward priority queue.
 * The elements are lexicographically sorted, first on x, then on y.
 * A non standard operation is the possibility to remove an arbitrary element.
 * In order to make this possible an identifier is returned at each insert.
 * This identifier can be used to remove this item from the queue.
 */

struct queue {
	    queue() {_top = 0;}
    qnode   *pop();
    queueid insert(qnode *);
    void    remove(queueid);
    bool    is_empty() const;
protected:
    qnode *_top;
    bool    is_consistent() const;
static void purge(qnode **curpt);
static void fill_hole(qnode **curpt, qnode *left, qnode *right);
static bool consistent_subtree(qnode **t);
};

// ************* implementation of union find *************

uf_handle uf_struct::new_handle()
{
    int result = _rr.size();
    _rr.grow(1);
    _rr[result] = result;
    return result;
}

void uf_struct::unite(uf_handle a, uf_handle b)
{
    assert(is_valid_handle(a));
    assert(is_valid_handle(b));
    _rr[find(b)] = find(a);
}

uf_handle uf_struct::find(uf_handle a) const
{
    assert(is_valid_handle(a));
    int next, cur, result;
    cur = a;
    while (1) {
	result = _rr[cur];
	if (result == cur)
	    break;
	cur = result;
    }
// path compression
    cur = a;
    while (cur != result) {
	next = _rr[cur];
	_rr[cur] = result;
//	((intreray &)_rr)[cur] = result;    // cast away const
	cur = next;
    }
    return result;
}

bool uf_struct::is_valid_handle(uf_handle h) const
{
    return h>= 0 && h<_rr.size();
}

bool uf_struct::is_equivalence_class(int i) const
{
    return i>=0 && i<_rr.size() && _rr[i] == i;
}

void uf_struct::all_ids(intreray &ids) const
// delivers all ids that can be returned by find sorted in increasing order.
{
    int i;
    ids.newsize(0);
    for (i=0; i<_rr.size(); i++)
	if (_rr[i] == i)
	    ids.append(i);
}

void region_uf_struct::associate(uf_handle cellh,
		dcel_cell_id cell, uf_handle comph)
{
    assert(is_valid_handle(cellh));
    int i;
    i = find(cellh);
    if (i >= _cell_id.size()) {
	_cell_id.resize(i+1);
	_outer_comp.resize(i+1);
    }
    _cell_id[i] = cell.index;
    _outer_comp[i] = comph;
}

dcel_cell_id region_uf_struct::get_cell_id(int i) const
{
    assert(is_equivalence_class(i));
    dcel_cell_id result;
    result.index = _cell_id[i];
    return result;
}

uf_handle region_uf_struct::get_surrounding_comp(int i) const
{
    assert(is_equivalence_class(i));
    int size;
    size = _outer_comp.size();
    return _outer_comp[i];
}


void component_uf_struct::associate(uf_handle comph,
			    dcel_edge_id edge, uf_handle cellh)
{
    assert(is_valid_handle(comph));
    int i;
    i = find(comph);
    if (i >= _outer_edge.size()) {
	_outer_edge.resize(i+1);
	_outer_cell.resize(i+1);
    }
    _outer_edge[i] = edge.index;
    _outer_cell[i] = cellh;
}

dcel_edge_id component_uf_struct::get_outer_edge(int i) const
{
    assert(is_equivalence_class(i));
    dcel_edge_id result;
    result.index = _outer_edge[i];
    return result;
}

uf_handle component_uf_struct::get_surrounding_cell(int i) const
{
    assert(is_equivalence_class(i));
    return _outer_cell[i];
}


// ************* implementation of event queue *************

void queue::remove(queueid qid)
{
    qid->_deleted = 1;
    debug_assert(is_consistent());
}

void queue::fill_hole(qnode **curpt, qnode *left, qnode *right)
{
    qnode *tmp;
    while (1) {
	purge(&left);
	purge(&right);
	if (left == 0) {
	    *curpt = right;
	    break;
	}
	if (right == 0) {
	    *curpt = left;
	    break;
	} 
	if (left->_vt <= right->_vt) {
	    tmp = left;
	    *curpt = left;
	    left = tmp->right;
	    tmp->right = right;
	    right = tmp->left;
	    curpt = &(tmp->left);
	} else {
	    tmp = right;
	    *curpt = right;
	    right = tmp->left;
	    tmp->left = left;
	    left = tmp->right;
	    curpt = &(tmp->right);
	}
    }
}

void queue::purge(qnode **curpt)
{
    qnode *cur = *curpt;
    if (cur == 0  || !cur->_deleted)
	return;
    fill_hole(curpt, cur->left, cur->right);
    delete cur;
}

queueid queue::insert(qnode *newnode)
{
    qnode *cur, *qn, **curpt;
    qn = newnode;
    curpt = &_top;
    purge(curpt);
    cur = *curpt;
    while (cur != 0) {
	if (qn->_vt <= cur->_vt) {
	    *curpt = qn;
	    qn->left = cur->left;
	    qn->right = cur->right;
	    qn = cur;
	    cur = *curpt;
	}
	if (rand()%2 == 0)
	    curpt = &cur->left;
	else
	    curpt = &cur->right;
	purge(curpt);
	cur = *curpt;
    }
    *curpt = qn;
    qn->left = qn->right = 0;
    debug_assert(is_consistent());
    return newnode;
}

qnode *queue::pop()
{
    qnode *result;
    purge(&_top);
    result = _top;
    if (_top == 0)
	return result;
    fill_hole(&_top, _top->left, _top->right);
    debug_assert(is_consistent());
    return result;
}

bool queue::is_empty() const
{
    purge((qnode **)(&_top));
    return _top == 0;
}

bool queue::consistent_subtree(qnode **t)
{
    purge(t);
    qnode *nd = *t;
    if (*t == 0)
	return true;
    if (!consistent_subtree(&nd->left))
	return false;
    if (!consistent_subtree(&nd->right))
	return false;
    if (nd->left != 0 && nd->left->_vt < nd->_vt)
	return false;
    if (nd->right != 0 && nd->right->_vt < nd->_vt)
	return false;
    return true;
}

bool queue::is_consistent() const
{
    return consistent_subtree((qnode **)(&_top));
}

// ****************************************


static void chain_start(dcedgelist &dcel,
	region_uf_struct &regions, component_uf_struct &components,
	const pl_vec &_vt, treeinfo *high, treeinfo *low)
// start two chains of dcel edges.
{
    dcel_vertex_id vertex;
    vertex = dcel.new_vertex();
    dcel.set_coordinates(vertex, _vt);
    dcel.set_info(vertex, (void *) components.new_handle());	// component
    high->upedge = dcel.new_edge();
    high->downedge = dcel.new_edge();
    low->upedge = dcel.new_edge();
    low->downedge = dcel.new_edge();
    dcel.set_origin(high->upedge, vertex);
    dcel.set_origin(low->upedge, vertex);
    dcel.make_twins(high->upedge, high->downedge);
    dcel.make_twins(low->upedge, low->downedge);
    dcel.make_consecutive(high->downedge, low->upedge);
    dcel.make_consecutive(low->downedge, high->upedge);
    high->_region = regions.new_handle();
}

static uf_handle component_handle(treeinfo *info, const dcedgelist &dcel)
{
    dcel_vertex_id vt;
    vt = dcel.origin(info->upedge);
    return (uf_handle) (dcel.info(vt));
}

static void chain_bend(dcedgelist &dcel, const pl_vec &_vt,
    treeinfo *info, treeinfo *updinfo)
// continue a chain of dcel edges.
{
    debug_assert(info->upedge != No_dcel_edge_id);
    debug_assert(info->downedge != No_dcel_edge_id);
    dcel_vertex_id vertex;
    dcel_edge_id newup, newdown;
    vertex = dcel.new_vertex();
    dcel.set_coordinates(vertex, _vt);
    dcel.set_info(vertex, (void *)component_handle(info, dcel));
    newup = dcel.new_edge();
    newdown = dcel.new_edge();
    dcel.set_origin(info->downedge, vertex);
    dcel.set_origin(newup, vertex);
    dcel.make_twins(newup, newdown);
    dcel.make_consecutive(info->upedge, newup);
    dcel.make_consecutive(newdown, info->downedge);
    info->upedge = No_dcel_edge_id;
    info->downedge = No_dcel_edge_id;
    updinfo->upedge = newup;
    updinfo->downedge = newdown;
}

static dcel_cell_id make_cell(dcedgelist &dcel,
		    dcel_edge_id outeredge, int cell_info)
{
    dcel_edge_id cur;
    dcel_cell_id cell;
    cell = dcel.new_cell();
    // set incident edges.
    debug_assert(cell != No_dcel_cell_id);
    cur = outeredge;
    do {
	dcel.set_incident_cell(cur, cell);
	cur = dcel.following(cur);
    } while (cur != outeredge);
    dcel.set_outside_edge(cell, outeredge);
    dcel.set_info(cell, (void *) cell_info);
    return cell;
}

static void chain_end(dcedgelist &dcel,
	region_uf_struct &regions, component_uf_struct &components,
	const pl_vec &_vt,
	treeinfo *info3, treeinfo *info2, treeinfo *info1, int cell_info)
// End two chains of dcel edges in a common vertex.
{
    dcel_vertex_id vertex;
    debug_assert(info1->upedge != No_dcel_edge_id);
    debug_assert(info1->downedge != No_dcel_edge_id);
    debug_assert(info2->upedge != No_dcel_edge_id);
    debug_assert(info2->downedge != No_dcel_edge_id);
    vertex = dcel.new_vertex();
    dcel.set_coordinates(vertex, _vt);
    dcel.set_origin(info2->downedge, vertex);
    dcel.set_origin(info1->downedge, vertex);
    dcel.make_consecutive(info1->upedge, info2->downedge);
    dcel.make_consecutive(info2->upedge, info1->downedge);
    regions.unite(info3->_region, info1->_region);
    uf_handle comp1, comp2;
    comp1 = component_handle(info1, dcel);
    comp2 = component_handle(info2, dcel);
    dcel.set_info(vertex, (void *) comp1);
    if (components.find(comp1) == components.find(comp2)) {
        regions.associate(info2->_region,
		  make_cell(dcel, info2->downedge, cell_info), comp1);
	info2->_region = INVALID_HANDLE;
    } else {
	components.unite(comp1, comp2);
    }
    components.associate(comp1, info2->upedge, info3->_region); // dit is outer edge van comp
    info2->upedge = No_dcel_edge_id;
    info2->downedge = No_dcel_edge_id;
    info1->upedge = No_dcel_edge_id;
    info1->downedge = No_dcel_edge_id;
}

static void chain_cross(dcedgelist &dcel,
	region_uf_struct &regions, component_uf_struct &components,
	const pl_vec &_vt, treeinfo *info2, treeinfo *info1, int cell_info)
// Four chains of dcel edges come togehter in one vertex.
{
    dcel_vertex_id vertex;
    dcel_edge_id newhighup, newhighdown, newlowup, newlowdown;
    vertex = dcel.new_vertex();
    newhighup = dcel.new_edge();
    newhighdown = dcel.new_edge();
    newlowup = dcel.new_edge();
    newlowdown = dcel.new_edge();

    dcel.set_coordinates(vertex, _vt);
    dcel.set_origin(info2->downedge, vertex);
    dcel.set_origin(info1->downedge, vertex);
    dcel.set_origin(newhighup, vertex);
    dcel.set_origin(newlowup, vertex);
    dcel.make_consecutive(info1->upedge, info2->downedge);
    dcel.make_consecutive(newhighdown, newlowup);
    dcel.make_consecutive(newlowdown, info1->downedge);
    dcel.make_consecutive(info2->upedge, newhighup);
    dcel.make_twins(newhighup, newhighdown);
    dcel.make_twins(newlowup, newlowdown);
    uf_handle comp1, comp2;
    comp1 = component_handle(info1, dcel);
    comp2 = component_handle(info2, dcel);
    dcel.set_info(vertex, (void *) comp1);
    if (components.find(comp1) == components.find(comp2)) {
        regions.associate(info2->_region,
		      make_cell(dcel, info2->downedge, cell_info), comp1);
    } else {
	components.unite(comp1, comp2);
    }
    info2->_region = info1->_region;
    info1->_region = regions.new_handle();
    info2->upedge = newlowup;
    info2->downedge = newlowdown;
    info1->upedge = newhighup;
    info1->downedge = newhighdown;
}

// ****************************************

struct event: public qnode {
	    event(const pl_vec &vt) : qnode(vt) {}
    virtual ~event();
    virtual void treat_event(region_uf_struct &, component_uf_struct &,
		    pl_sweeplist &, queue &,
		    dcedgelist &, int (*)(int)) = 0;
};

event::~event() {}

struct insert_event: public event {
		    insert_event(const pl_vec &vt) : event(vt){}
    void treat_event(region_uf_struct &, component_uf_struct &,
			pl_sweeplist &, queue &,
			dcedgelist &, int (*)(int));
    bool	    _reflex;	// Marks start or end of polygon.
    pl_linereray    _highedges;	// The high chain of edges starting in _vt.
    vtident	    _highend;	// Identifier of the last vertex on high chain.
    pl_linereray    _lowedges;	// The low chain of edges starting in _vt.
    vtident	    _lowend;	// Identifier of the last vertex on low chain.
};


struct transition_event: public event {
    transition_event(const pl_vec &vt, pl_sweep_item it) : event(vt) {_it = it;}
    void treat_event(region_uf_struct &, component_uf_struct &,
			pl_sweeplist &, queue &,
			dcedgelist &, int (*)(int));
protected:
    pl_sweep_item _it;		// Associated leaf in status tree.
};

struct delete_event: public event {
    delete_event(const pl_vec &vt, pl_sweep_item it) : event(vt) {_it = it;}
    void treat_event(region_uf_struct &, component_uf_struct &,
			pl_sweeplist &, queue &,
			dcedgelist &, int (*)(int));
protected:
    pl_sweep_item _it;		// Associated leaf in status tree.
};

struct cross_event: public event {
    cross_event(const pl_vec &vt, pl_sweep_item it) : event(vt) {_it = it;}
    void treat_event(region_uf_struct &, component_uf_struct &,
			pl_sweeplist &, queue &,
			dcedgelist &, int (*)(int));
protected:
    pl_sweep_item _it;		// Associated leaf in status tree.
};


// ****************************************

static int checkintersect(pl_vec &isvt, treeinfo *highinfo, treeinfo *lowinfo)
{
    // check altijd vertex van lijn met kleinste end (vtident) met
    // lijn grootste end (vanwege robustness)
    if (highinfo->at_end() && lowinfo->at_end() && highinfo->end == lowinfo->end)
	return 0;
    const pl_line &highline = highinfo->edges[highinfo->edgeindex];
    const pl_line &lowline = lowinfo->edges[lowinfo->edgeindex];
    const pl_vec &highvt = highline.ref_vt();
    const pl_vec &lowvt = lowline.ref_vt();
    if (highvt < lowvt) {
	if (is_to_the_left(highvt, lowline))
	    return 0;
	if(intersect(isvt, highline, lowline)) {
	    if (isvt > highvt)
		isvt = highvt; // ???????????
	} else {
	    isvt = highvt; // ???????????
	}
//	debug_assert(!isnan_pp(isvt.x()));
//	debug_assert(!isnan_pp(isvt.y()));
	return 1;
    } else {
    if (highvt > lowvt) {
	if (is_to_the_right(lowvt, highline))
	    return 0;
	if(intersect(isvt, highline, lowline)) {
	    if (isvt > lowvt)
		isvt = lowvt;
	} else {
	    isvt = lowvt; // ???????????
	}
//	debug_assert(!isnan_pp(isvt.x()));
//	debug_assert(!isnan_pp(isvt.y()));
	return 1;
    } else {
	return 0;
    }
    }
}

static void check_event(queue &q, treeinfo *highinfo, treeinfo *lowinfo, 
	    pl_sweep_item item)
{
    pl_vec isvt;
    event *evt;
    if ( highinfo->at_end() && lowinfo->at_end()
      && highinfo->end == lowinfo->end) {
	evt = new delete_event(lowinfo->end_vt(), item);
	highinfo->qid = q.insert(evt);
	return;
    }
    if (checkintersect(isvt, highinfo, lowinfo)) {
	evt = new cross_event(isvt, item);
	highinfo->qid = q.insert(evt);
	return;
    }
    highinfo->qid = no_event;
}

static void check_transition(queue &q, treeinfo *node, pl_sweep_item it)
{
    if (!node->at_end()) {
	event *evt;
	evt = new transition_event(node->edges[node->edgeindex].ref_vt(), it);
	q.insert(evt);
    }
}

static void remove_event(queue &q, treeinfo *info)
{
    if (!is_no_event(info->qid)) {
	q.remove(info->qid);
	info->qid = no_event;
    }
}

static bool consistent_border(region_uf_struct &regions,
		treeinfo *high, treeinfo *low)
{
    if (low->upedge == No_dcel_edge_id) {
	if (low->upedge != No_dcel_edge_id)
	    return false;
	if (regions.find(high->_region) != regions.find(low->_region))
	    return false;
	return true; 
    } else {
	if (low->upedge == No_dcel_edge_id)
	    return false;
	if (regions.find(high->_region) == regions.find(low->_region))
	    return false;
	return true; 
    }
}

void insert_event::treat_event(region_uf_struct &regions,
	component_uf_struct &components, pl_sweeplist &tree,
	queue &q, dcedgelist &dcel, int (*equi_class)(int))
{
    treeinfo *above, *high, *low;
    pl_sweep_item item0, item1, item2, item3;
    low = new treeinfo;
    high = new treeinfo;
    item1 = tree.find_and_insert(_vt, low);
    item2 = tree.insert_after(item1, high);
    item3 = tree.succ(item2);
    above = conv(tree.info(item3));
    high->end = _highend;
    high->edges.swap_contents(_highedges);
    high->edgeindex = 0;
    low->end = _lowend;
    low->edges.swap_contents(_lowedges);
    low->edgeindex = 0;
    low->_region = above->_region;
    low->_coverage = above->_coverage;
    high->_coverage = above->_coverage;
    low->qid = no_event;
    high->qid = no_event;
    if (_reflex) {
	high->orientation = LOWERBOUND;
	low->orientation = UPPERBOUND;
	high->_coverage--;
    } else {
	high->orientation = UPPERBOUND;
	low->orientation = LOWERBOUND;
	high->_coverage++;
    }
    // remove possible event between `above' and `below'.
    remove_event(q, above);
    // check of transitions aan queue toegevoegd moeten worden
    check_transition(q, high, item2);
    check_transition(q, low, item1);
    // check intersections van high line met above en low line met below
    if (!above->is_highest())
	check_event(q, above, high, item3);
    item0 = tree.pred(item1);
    if (item0 != 0)
	check_event(q, low, conv(tree.info(item0)), item1);
    int class1, class2;
    class1 = equi_class(above->_coverage);
    class2 = equi_class(high->_coverage);
    if (class1 != class2) {
	chain_start(dcel, regions, components, _vt, high, low);
    } else {
	high->upedge = No_dcel_edge_id;
	high->downedge = No_dcel_edge_id;
	low->upedge = No_dcel_edge_id;
	low->downedge = No_dcel_edge_id;
	high->_region = above->_region;
    }
    debug_assert(consistent_border(regions, above, high));
    debug_assert(consistent_border(regions, high, low));
}

static void assert_empty(queue &q, treeinfo *leaf)
{
    if (!is_no_event(leaf->qid)) {
	cerr<<"Inconsistency: leaf not empty.\n";
	q.remove(leaf->qid);
	leaf->qid = no_event;
    }
}

void transition_event::treat_event(region_uf_struct &regions,
		component_uf_struct &, pl_sweeplist &tree,
		queue &q, dcedgelist &dcel, int (*equi_class)(int))
{
    pl_sweep_item below_item, above_item;
    treeinfo *info, *above_info, *below_info;
    info = conv(tree.info(_it));
    below_item = tree.pred(_it);
    above_item = tree.succ(_it);
    if (below_item != 0) {
	below_info = conv(tree.info(below_item));
//	assert_empty(q, below_info);
    }
    above_info = conv(tree.info(above_item));
    if (!above_info->is_highest()) {
	assert_empty(q, above_info);
    }
    info->edgeindex++;
    check_transition(q, info, _it);
    if (!above_info->is_highest())
	check_event(q, above_info, info, above_item);
    if (below_item != 0)
	check_event(q, info, below_info, _it);
    if (info->upedge != No_dcel_edge_id) {
	debug_assert(info->downedge != No_dcel_edge_id);
	debug_assert(equi_class(above_info->_coverage) != equi_class(info->_coverage));
	chain_bend(dcel, _vt, info, info);
    } else {
	debug_assert(info->downedge == No_dcel_edge_id);
	debug_assert(equi_class(above_info->_coverage) == equi_class(info->_coverage));
    }
    debug_assert(consistent_border(regions, above_info, info));
}

void delete_event::treat_event(region_uf_struct &regions,
	component_uf_struct &components, pl_sweeplist &tree,
	queue &q, dcedgelist &dcel, int (*equi_class)(int))
{
    pl_sweep_item item0, item1, item2, item3;
    treeinfo *info0, *info1, *info2, *info3;
    item2 = _it;
    info2 = conv(tree.info(item2));
    item3 = tree.succ(item2);
    info3 = conv(tree.info(item3));
    item1 = tree.pred(item2);
    info1 = conv(tree.info(item1));
    item0 = tree.pred(item1);
    if (item0 != 0) {
	info0 = conv(tree.info(item0));
    }
    if (info2->upedge != No_dcel_edge_id) {
	chain_end(dcel, regions, components, _vt,
		    info3, info2, info1, equi_class(info2->_coverage));
    } else {
    }
// remove from tree
    remove_event(q, info1);
    remove_event(q, info3);
    tree.remove(item1);
    tree.remove(item2);
    delete info1;
    delete info2;
    // check if new neighbours intersect.
    assert_empty(q, info3);
    if (!info3->is_highest() && item0 != 0)
	check_event(q, info3, info0, item3);
}

static void treat_coverage(dcedgelist &dcel, region_uf_struct &regions,
    component_uf_struct &components, const pl_vec &vt, 
    treeinfo *info3, treeinfo *info2, treeinfo *info1, 
    int left, int right, int top, int bottom)
{
    if (top == bottom) {
	if (left == right) {
	    if (left != top) {
		chain_cross(dcel, regions, components, vt, info2, info1, left);
	    }
	} else {    // left != right, top == bottom
	    if (left == top) {
		chain_start(dcel, regions, components, vt, info1, info2);
	    } else {
		if (right == top) {
		chain_end(dcel, regions, components, vt,
				info3, info2, info1, left);
		info2->_region = info3->_region;
		info1->_region = info3->_region;
		} else {
		chain_cross(dcel, regions, components, vt, info2, info1, left);
		}
	    }
	}
    } else { // top != bottom
	debug_assert(right == left);
	if (left == top) {
	    chain_bend(dcel, vt, info1, info2);
	    info2->_region = info1->_region;
	    info1->_region = info3->_region;
	} else {    // left != top, top != bottom
	    if (left == bottom) {
		chain_bend(dcel, vt, info2, info1);
		info1->_region = info2->_region;
	    } else {
		chain_cross(dcel, regions, components, vt, info2, info1, left);
	    }
	}
    }
}

void cross_event::treat_event(region_uf_struct &regions,
	component_uf_struct &components, pl_sweeplist &tree,
	queue &q, dcedgelist &dcel, int (*equi_class)(int))
{
    pl_sweep_item item0, item1, item2, item3;
    treeinfo *info0, *info1, *info2, *info3;
    item2 = _it;
    info2 = conv(tree.info(item2));
    item3 = tree.succ(item2);
    info3 = conv(tree.info(item3));
    item1 = tree.pred(item2);
    info1 = conv(tree.info(item1));
    item0 = tree.pred(item1);
    if (item0 != 0)
	info0 = conv(tree.info(item0));
    // _coverage aanpassen en eventueel nieuwe chain starten
    int leftcov,  rightcov, topcov, bottomcov;
    leftcov = info2->_coverage;
    topcov = info3->_coverage;
    bottomcov = info1->_coverage;
    rightcov = leftcov + (info1->orientation - info2->orientation); // ?????
    treat_coverage(dcel, regions, components, _vt, 
		info3, info2, info1, 
		equi_class(leftcov),  equi_class(rightcov),
		equi_class(topcov), equi_class(bottomcov));
    info2->_coverage = bottomcov;
    info1->_coverage = rightcov;
    // verwissel in tree
    tree.exchange(item1, item2);
    // examine if new intersections emerge.
    remove_event(q, info1);
    remove_event(q, info2);
    if (item0 != 0)
	check_event(q, info2, info0, item2);
    remove_event(q, info3);
    if (!info3->is_highest())
	check_event(q, info3, info1, item3);
    debug_assert(consistent_border(regions, info3, info1));
    debug_assert(consistent_border(regions, info1, info2));
}


// ****************************************

static int upward_vertex(const pl_pgn &pgn, int i)
{
    const pl_vec &vti = pgn.vertex(i);
    return (cross(pgn.vertex_mod(i-1) - vti, pgn.vertex_mod(i+1) - vti) > 0);
}

static void add_starts(queue &q, const pl_pgn &pgn, int add)
{
    static intreray visited;
    int i, order, clockwise;
    insert_event *evt;
    if (pgn.size() <= 2)
	return;
    clockwise = pgn.is_clockwise();
    visited.newsize(pgn.size());
    for (i=0; i<visited.size(); i++)
	visited[i] = 0;
    for (order=0; order<pgn.size(); order++) {
	i = pgn.element(order);
	if (visited[i])
	    continue;
	visited[i] = 1;
	evt = new insert_event(pgn.vertex(i));
	// bepaal of i+1 of i-1 hoger is (cross) ->reflex of niet
	pl_linereray *positive_edges, *negative_edges;
	int *positive_end, *negative_end;
	if (upward_vertex(pgn, i)) {
	    positive_edges = &evt->_highedges;
	    negative_edges = &evt->_lowedges;
	    positive_end = &evt->_highend;
	    negative_end = &evt->_lowend;
	    evt->_reflex = !clockwise ;
	} else {
	    negative_edges = &evt->_highedges;
	    positive_edges = &evt->_lowedges;
	    negative_end = &evt->_highend;
	    positive_end = &evt->_lowend;
	    evt->_reflex = clockwise;
	}
	// make positive edges
	int j, next;
	j = i;
	while (1) {
	    next = j + 1;
	    if (next >= pgn.size())
		next = 0;
	    if (pgn.order(next) < pgn.order(j))
		break;
	    visited[next] = 1;
	    const pl_vec &endvt = pgn.vertex(next);
	    positive_edges->append(pl_line(endvt, normalised(endvt-pgn.vertex(j))));
	    j = next;
	}
	*positive_end = j+add;
	// make negative edges
	j = i;
	while (1) {
	    next = j - 1;
	    if (next < 0 )
		next = pgn.size()-1;
	    if (pgn.order(next) < pgn.order(j))
		break;
	    visited[next] = 1;
	    const pl_vec &endvt = pgn.vertex(next);
	    negative_edges->append(pl_line(endvt, normalised(endvt-pgn.vertex(j))));
	    j = next;
	}
	*negative_end = j+add;
	q.insert(evt);
    }
}

static int binsearch(const intreray &irr, uf_handle region, uf_struct &regions)
{
    int s, low, high, mid, cmp;
    s = regions.find(region);
    low = 0;
    high = irr.size();
    while (1) {
	mid = (low+high)/2;
	cmp = irr[mid];
	if (cmp <s) {
	    low = mid+1;
	} else {
	if (cmp > s) {
	    high = mid;
	} else {
	    break;
	}
	}
    }
    return mid;
}

static void complete_dcel(dcedgelist &dcel, region_uf_struct &regions,
	component_uf_struct &components)
{
    intreray allregions, allcomponents;
    int i;
    components.all_ids(allcomponents);
// create all components
    for (i=0; i<allcomponents.size(); i++)
	dcel.new_component();
    dcel. update_int_idents();
// convert all uf_handles of the vertices to dcel_comp_id.
    dcel_vertex_id vertex;
    for (vertex=dcel.first_vertex();
	    vertex != No_dcel_vertex_id;
	    vertex = dcel.next_vertex(vertex)) {
        int compintid;
	compintid = binsearch(allcomponents,
		    (uf_handle) dcel.info(vertex), components);
	dcel.set_component(vertex, dcel.comp_id(compintid));
	dcel.set_info(vertex, 0);
    }
// set surrounding cell of every component and
// set incident_cell of every outer edge of components.
    for (i=0; i < allcomponents.size(); i++) {
	uf_handle cellh;
	int comph = allcomponents[i];
	dcel_cell_id cell;
	dcel_edge_id first, cur;
	cellh = components.get_surrounding_cell(comph);
	cell = regions.get_cell_id(regions.find(cellh));
	debug_assert(cell != No_dcel_cell_id);
	dcel.set_surrounding_cell(dcel.comp_id(i), cell);
	first = components.get_outer_edge(comph);
	dcel.set_outside_edge(dcel.comp_id(i), first);
	cur = first;
	do {
	    debug_assert(dcel.incident_cell(cur) == No_dcel_cell_id);
	    dcel.set_incident_cell(cur, cell);
	    cur = dcel.following(cur);
	} while (cur != first);
    }
// set surrounding component of every cell.
    regions.all_ids(allregions);
    for (i=0; i<allregions.size(); i++) {
	dcel_cell_id cell;
        int compintid;
	uf_handle comph;
	int regioni = allregions[i];
	comph = regions.get_surrounding_comp(regioni);
	if (comph != INVALID_HANDLE) {
	    cell = regions.get_cell_id(allregions[i]);
	    compintid = binsearch(allcomponents, comph, components);
	    dcel_comp_id comp = dcel.comp_id(compintid);
	    debug_assert(comp != No_dcel_comp_id);
	    dcel.set_surrounding_component(cell, comp);
	}
    }
}

void sweep_phase(dcedgelist &dcel, region_uf_struct &regions,
	component_uf_struct &components,
	const reray<pl_pgn> &in, int (*equi_class)(int))
{
    queue	    q;		    // The event queue.
    pl_sweeplist    tree(tree_cmp); // The status tree.
    treeinfo	    *highest;	    // The highest leaf of the status tree.
    uf_handle	    startregion;    // The outside region.
    int		    i;
    int		    idadd;	    // Used to assign a unique ident to
				    // each end vertex.

    idadd = 0;
    dcel.clear();
    startregion = regions.new_handle();

    #ifdef VISUALISE
	pl_box bbox;
	pl_glwindow drawarea;
	bbox.make_empty();
    #endif

// Add all start vertices of all polygons to the queue.
    for (i=0; i<in.size(); i++) {
	const pl_pgn &curpgn = in[i];
	add_starts(q, curpgn, idadd);
	idadd += curpgn.size();

	#ifdef VISUALISE
	    bbox += curpgn.bbox();
	    drawarea.set_viewingarea(bbox.xmin()-1, bbox.ymin()-1, 
		    bbox.xmax()-bbox.xmin()+2, bbox.ymax()-bbox.ymin()+2);
	#endif

    }

// Insert a symbolical highest chain in the status tree.
    highest = new treeinfo;
    highest->end = -1;			// This signals this special node.
    highest->qid = no_event;
    highest->upedge = No_dcel_edge_id;
    highest->downedge = No_dcel_edge_id;
    highest->_coverage = 0;		// The coverage of the outside region.
    highest->_region = startregion;	// The outside region.
    tree.insert_first(highest);
// Treat all events in the queue. New events are generated during this.
    while (!q.is_empty()) {
	event *curevent;
	curevent = (event *)(q.pop());

	#ifdef VISUALISE
	    drawarea.set_linecol(0);
	    for (i=0; i<in.size(); i++) {
		drawarea.draw(in[i]);
	    }
	    drawarea.set_linecol(4);
	    drawarea.draw(pl_line(curevent->_vt, pl_ydir));
	    drawarea.set_linecol(1);
	    drawarea.draw(curevent->_vt);
	    drawarea.pageready();
	    while (!drawarea.get_mousestatus(3))
		;
	    while (drawarea.get_mousestatus(3))
		;
	#endif

	curevent->treat_event(regions, components, tree, q, dcel, equi_class);
	delete curevent;
    }
    regions.associate(startregion, dcel.outer_cell(), INVALID_HANDLE);
    delete highest;
}

/*
 * pgncoverage takes a reray of polygons and makes a doubly connected edgelist.
 * This dcel will divide the space in regions depending on by how many
 * polygons a region is covered.
 * The algorithm is divided in two phases. The first phase is a plane sweep.
 * In this phase the edges, vertices and cells of the dcel are created.
 * Furthermore information about the extent of the cells and components is
 * gathered in a union-find structure.
 * In the second phase the information that is collected is used to complete
 * the dcel. The components are created and the is_surrounded_by is made.
 */
void pgncoverage(dcedgelist &dcel, const reray<pl_pgn> &in,
    int (*equi_class)(int))
{
    region_uf_struct regions;
    component_uf_struct components;
    sweep_phase(dcel, regions, components, in, equi_class);
    // fix final incidence relations (surrounding components)
    complete_dcel(dcel, regions, components);
}

static int at_least_one(int c)
{
    return c > 0 ? 1 : 0;
}

void pgnunion(dcedgelist &dcel, const reray<pl_pgn> &in)
{
    pgncoverage(dcel, in, at_least_one);
}
