// $Id: geolib.C,v 1.3 1995/10/19 10:51:03 geert Exp $

#include <PLAGEO/pgn.h>
#include <PLAGEO/clearance1.h>
#include <geolib.h>

//////////////////////////////////////////////////
//						//
//  GeoLib					//
//	A library of higher level routines	//
//	based on plageo				//
//						//
//////////////////////////////////////////////////





//////////////////////////////////////////////////////////////
//							    //
//  COMPUTING CONVEX HULLS				    //
//							    //
//////////////////////////////////////////////////////////////

static void find_hull(pl_vecreray &vert, pl_vec from, pl_vec to)
// Computes the chain of vertices on the convex hull from from to to
{
    int i;

    //// Check easy cases
    if (vert.size() <= 1) return;
    
    //// Find vertex furthest away
    pl_line line(from, to);		    // Line from from to to
    int ind = 0;			    // Index of furthest point
    float dist = clearance(vert[0], line);  // Best distance so far
    float dd;				    // Temporary distance
    for (i=1; i<vert.size(); i++) {
	dd = clearance(vert[i], line);
	if (dd > dist) {dist = dd; ind = i; }
    }
    
    //// Compute the two sets
    pl_vecreray first, second;	    // Sets of remaining points
    line = pl_line(from, vert[ind]);
    for (i=0; i<vert.size(); i++)
	if (i!=ind && is_to_the_right(vert[i], line))
	    first.append(vert[i]);
    line = pl_line(vert[ind], to);
    for (i=0; i<vert.size(); i++)
	if (i!=ind && is_to_the_right(vert[i], line))
	    second.append(vert[i]);
	    
    //// Recursive calls and merge
    pl_vec middle = vert[ind];	    // Middle vertex
    find_hull(first, from, middle);
    find_hull(second, middle, to);
    vert.newsize(first.size()+1+second.size());
    for (i=0; i<first.size(); i++) vert[i] = first[i];
    vert[first.size()] = middle;
    for (i=0; i<second.size(); i++) vert[first.size()+i+1] = second[i];
}


pl_pgn convex_hull(const pl_vecreray &vert)
// Computes the convex hull of the set of vertices using quickhull
{
    int i;
    
    //// Do the easy cases
    if (vert.size() <= 3) return pl_pgn(vert);
    
    //// Compute leftmost and rightmost vertex
    int left = 0,  right = 0;		// Index of leftmost and rightmost
    for (i=1; i<vert.size(); i++)
    {
	if (vert[i]< vert[left]) left = i;
	if (vert[i] >= vert[right]) right = i;
    }
    
    //// Split into left and right half
    pl_line line(vert[left], vert[right]);  // Line through the two points
    pl_vecreray leftvert, rightvert;	    // Vertices left and right of the line
    for (i=0; i<vert.size(); i++)
	if (i!=left && i!=right) {
	    if (is_to_the_left(vert[i], line))
		leftvert.append(vert[i]);
	    else
		rightvert.append(vert[i]);
	}
    
    //// Recursive calls and merging
    find_hull(leftvert, vert[right], vert[left]);
    find_hull(rightvert, vert[left], vert[right]);
    pl_vecreray result;		    // Final list of vertices
    result.newsize(leftvert.size() + rightvert.size() + 2);
    result[0] = vert[right];
    for (i=0; i<leftvert.size(); i++)
	result[i+1] = leftvert[i];
    int base = 2 + leftvert.size();
    result[base-1] = vert[left];
    for (i=0; i<rightvert.size(); i++)
	result[base+i] = rightvert[i];
    return pl_pgn(result);
}

pl_pgn convex_hull(const pl_pgn &pol)
// Computes the convex hull of the polygon using quickhull on the vertices
{
    pl_vecreray vert;
    pol.all_vertices(vert);
    return convex_hull(vert);
}



//////////////////////////////////////////////////////////////
//							    //
//  COMPUTING MINKOWSKI SUMS				    //
//							    //
//////////////////////////////////////////////////////////////

pl_pgn Minkowski_sum(const pl_pgn &pol1, const pl_pgn &pol2)
// Computes the Minkowski sum of the two polygons. Only defined
// when both are convex. Otherwise the Minkowski sums of the
// convex hulls is computed.
{
    //// Compute a list of all sums of the vertices
    pl_vecreray vert(pol1.size() * pol2.size());
    int i, j, k, size1, size2;
    size1 = pol1.size();
    size2 = pol2.size();
    k = 0;
    for (i=0; i<size1; i++) {
	const pl_vec &vt1 = pol1.vertex(i);
	for (j=0; j<size2; j++)
	  vert[k++] = vt1 + pol2.vertex(j);
    }
    //// Return the convex hull of this set
    return convex_hull(vert);
}

void Minkowski_sum(reray<pl_pgn> &result, const pl_pgn &pol1, const pl_pgn &pol2)
// Computes the Minkowski sum of the two polygons. A set of convex polygons
// is returned that together cover the Minkowski sum.
{
    //// Split up the polygons in convex pieces
    reray<pl_pgn> first, second;
    convex_pieces(first, pol1);
    convex_pieces(second, pol2);
    
    //// For each pair compute Minkowski sum
    result.newsize(first.size() * second.size());
    int i, j, k;
    k = 0;
    for (i=0; i<first.size(); i++)
	for (j=0; j<second.size(); j++)
	    result[k++] = Minkowski_sum(first[i], second[j]);
}
