#ifndef PLAGEO_BASE_H
#define PLAGEO_BASE_H

#include <math++.h>
#include <reray.h>

#include <PLAGEO/basebase.h>

/*******************************
 *
 * basic types (classes) for planar geometry
 *
 * angles
 * vectors
 * bounding boxes
 *
 *******************************/



typedef reray<int> intreray;

inline bool increasing_angle_order(pl_angle small, pl_angle theta, pl_angle large)
{ return theta.lies_between(small, large); }
// returns true if, travelling over the unit circle counterclockwise, you will
// encounter the angles small, theta and large in that order.
// returns false if you would encounter them in the order large, theta, small.

inline bool decreasing_angle_order(pl_angle large, pl_angle theta, pl_angle small)
{ return theta.lies_between(small, large); }
// opposite of increasing_angle_order


// vectors


class pl_fixedvec {
public:
		    pl_fixedvec()	{}
		    pl_fixedvec(s_coord x, s_coord y)  { c[0]=x; c[1]=y;}
    const s_coord   &x() const			    {return c[0];}
    const s_coord   &y() const			    {return c[1];}
    s_coord	    length() const	{ return sqrt_pp(c[0]*c[0] + c[1]*c[1]); }
    s_coord	    sqlength() const	{ return c[0]*c[0] + c[1]*c[1]; }
protected:
    s_coord c[2];
private:
    void	    operator=(const pl_fixedvec &) ;	// not implemented.
};

inline bool operator< (const pl_fixedvec &a, const pl_fixedvec &b)
{
    bool result;
    if (a.x() == b.x()) result = (a.y() < b.y());
    else result = (a.x() < b.x());
    return result;
}

inline bool operator> (const pl_fixedvec &a, const pl_fixedvec &b)
{
    bool result;
    if (a.x() == b.x()) result = (a.y() > b.y());
    else result = (a.x() > b.x());
    return result;
}

inline bool operator<= (const pl_fixedvec &a, const pl_fixedvec &b)
{
    bool result;
    if (a.x() == b.x()) result = (a.y() <= b.y());
    else result = (a.x() < b.x());
    return result;
}

inline bool operator>= (const pl_fixedvec &a, const pl_fixedvec &b)
{
    bool result;
    if (a.x() == b.x()) result = (a.y() >= b.y());
    else result = (a.x() > b.x());
    return result;
}

extern int compare(const pl_fixedvec &a, const pl_fixedvec &b);

class pl_vec: public pl_fixedvec {
public:
		pl_vec()						    {}
		pl_vec(const pl_fixedvec &vec)		: pl_fixedvec(vec) {}
		pl_vec(const s_coord x, const s_coord y) : pl_fixedvec(x, y) {}
    void	normalise()		{s_coord l=1/length();c[0]*=l;c[1]*=l;}
    void	normalize()		{ normalise();}
    void        set_x(s_coord x)				{ c[0] = x;}
    void        set_y(s_coord y)				{ c[1] = y;}
    void	operator=(const pl_fixedvec &vec)
			    { c[0] = vec.x(); c[1] = vec.y(); }
    void	operator=(const pl_vec &vec)
			    { c[0] = vec.x(); c[1] = vec.y(); }
    void	operator*=(s_coord f)		{ c[0] *= f; c[1] *= f;}
    void	operator+=(const pl_fixedvec &vec)
		    {c[0] += vec.x(); c[1] += vec.y(); }
    void	operator-=(const pl_fixedvec &vec)
		    {c[0] -= vec.x(); c[1] -= vec.y(); }
};


class pl_unitvec: public pl_fixedvec {
// assert: length == 1
public:
		    pl_unitvec(): pl_fixedvec(1.0, 0.0)	{}
		    pl_unitvec(pl_angle angle) ;
    pl_unitvec	    operator-() const	{ return pl_unitvec(-x(), -y()); }
    pl_unitvec	    normal() const	    { return pl_unitvec(-y(), x());}
friend pl_unitvec   normalised(const pl_fixedvec &vec)	;
friend s_coord	    factorise(pl_unitvec &uv,const pl_fixedvec &vec);
    void	    operator=(const pl_unitvec &uvec)
				    { c[0]=uvec.x(); c[1]=uvec.y(); }
protected:   
		pl_unitvec(const s_coord a, const s_coord b): pl_fixedvec(a, b){}
		pl_unitvec(const pl_fixedvec &vec): pl_fixedvec(vec) {}
};

inline pl_unitvec normalized(const pl_fixedvec &vec)
{ return normalised(vec); }

inline s_coord factorize(pl_unitvec &uvec,const pl_fixedvec &vec)
{ return factorise(uvec, vec); }

const pl_vec pl_origin(0, 0);
extern const pl_unitvec pl_xdir;
extern const pl_unitvec pl_ydir;

extern pl_angle angle_of(const pl_fixedvec &vec);

pl_vec to_polar(const pl_fixedvec &);

pl_vec from_polar(const pl_fixedvec &);

extern pl_angle angle_between(const pl_unitvec &, const pl_unitvec &);


inline pl_vec operator- (const pl_fixedvec &vec)
{
    return pl_vec(-vec.x(), -vec.y());
}

inline pl_vec operator*(s_coord f, const pl_fixedvec &v)
{
    return pl_vec(f*v.x(), f*v.y());
}

inline pl_vec operator+ (const pl_fixedvec &v1, const pl_fixedvec &v2)
{
    return pl_vec(v1.x()+v2.x(), v1.y()+v2.y());
}

inline pl_vec operator- (const pl_fixedvec &v1, const pl_fixedvec &v2)
{
    return pl_vec(v1.x()-v2.x(), v1.y()-v2.y());
}

inline s_coord dot(const pl_fixedvec &v1, const pl_fixedvec &v2)
{
    return v1.x()*v2.x() + v1.y()*v2.y();
}

inline s_coord cross(const pl_fixedvec &v1, const pl_fixedvec &v2)
{
    return v1.x()*v2.y() - v1.y()*v2.x();
}

class pl_boundingbox {
public:
		pl_boundingbox()	{}
		pl_boundingbox(pl_fixedvec c) :
			min(c),max(c){}
    bool	is_empty() const	{return min.x() > max.x();}
    s_coord 	xmax() const		{return max.x();}
    s_coord 	xmin() const		{return min.x();}
    s_coord 	ymax() const		{return max.y();}
    s_coord 	ymin() const		{return min.y();}
    void	make_empty()		{min.set_x(1); max.set_x(-1);}
    s_coord	area() const		;
    void	translate(const pl_fixedvec &disp);
    void	scale(s_coord fac);
    void	add_point(const pl_fixedvec &c);
    void	add_first_point(const pl_fixedvec &c) {min = c; max = c; }
    void	add_non_first_point(const pl_fixedvec &c) ;
    void	operator+=(const pl_boundingbox &);
    friend pl_boundingbox 
		intersect(pl_boundingbox const &, pl_boundingbox const &);
    friend pl_boundingbox 
		operator+(const pl_boundingbox &, const pl_boundingbox &);
protected:
    pl_vec    min;
    pl_vec    max;
};

typedef pl_boundingbox pl_box;

extern bool lies_inside(const pl_box &innerbb, const pl_box &outerbb);

class pl_transform {
public:
    virtual pl_vec	operator*(const pl_vec &) const =0;
    virtual pl_unitvec	operator*(const pl_unitvec &) const =0;
    virtual s_coord	operator*(s_coord) const =0;
    virtual pl_angle	operator*(pl_angle) const = 0;
};

enum pl_object_type {PL_POINT=1, PL_LINE, PL_RAY, PL_EDGE,
		    PL_POLYGON, PL_DISC, PL_ARC};

class pl_object {
public:
    virtual	    ~pl_object() ;
    virtual pl_box  bbox() const  = 0;
    virtual void    transform(const pl_transform &tfm) = 0;
    virtual void    rotate(pl_angle) ;
    virtual void    translate(const pl_fixedvec &) ;
    virtual void    scale(s_coord) ;
    virtual pl_object_type type() const = 0;
};

class pl_point: public pl_object {
public:
		pl_point()						{}
		pl_point(const pl_fixedvec& v)			:_vec(v) {}
		pl_point(s_coord a, s_coord b) : _vec(a, b) {}
		~pl_point() ;
    pl_object_type type() const;
    pl_box	bbox() const ;
const pl_vec	&vec() const					{return _vec;}
    void	transform(const pl_transform &tfm);
    void	translate(const pl_fixedvec &disp) ;
    void	scale(s_coord fac);
protected:
    pl_vec	_vec;
};


typedef reray<pl_vec> pl_vecreray;

#endif // PLAGEO_BASE_H
