#include <PLAGEO/ipefile.h>
#include <PLAGEO/io.h>
#include <PLAGEO/coincidence.h>

const int IPESIZE = 400;


static void strokecol(ostream &out, int col)
{
    if (col<0 || col>=PL_COLORCOUNT) {
	strokecol(out, 0);
	return;
    }
    int red, green, blue;
    pl_color_components(red, green, blue, col);
    out << "% skc "  << red/255.0 << ' '
	<< green/255.0 << ' ' << blue/255.0 << '\n' ;
}

static void fillcol(ostream &out, int col)
{
    if (col<0 || col>=PL_COLORCOUNT) {
	fillcol(out, 0);
	return;
    }
    int red, green, blue;
    pl_color_components(red, green, blue, col);
    out << "% fic "  << red/255.0 << ' '
	<< green/255.0 << ' ' << blue/255.0 << '\n' ;
}

pl_box pl_ipefile::viewing_box() const
{
    pl_box bb;
    bb.add_first_point(pl_vec(_viewing_x, _viewing_y));
    bb.add_non_first_point(pl_vec(_viewing_x+_viewing_xsize,
				    _viewing_y+_viewing_ysize));
    return bb;
}

pl_vec pl_ipefile::to_ipe_coord(const pl_fixedvec &vt) const
{
    pl_vec result;
    float viewing_size = _viewing_xsize>_viewing_ysize ?
	    _viewing_xsize : _viewing_ysize;
    float fac = IPESIZE / viewing_size;
    result.set_x(fac * (vt.x() - _viewing_x - _viewing_xsize/2));
    result.set_y(fac * (vt.y() - _viewing_y - _viewing_ysize/2));
    return result;
}

s_coord pl_ipefile::to_ipe_size(s_coord x) const
{
    float viewing_size = _viewing_xsize>_viewing_ysize ?
	    _viewing_xsize : _viewing_ysize;
    return x * IPESIZE / viewing_size;
}


void pl_ipefile::draw(const pl_fixedvec &vt)
{
    page_init();
    if (!lies_inside(vt, viewing_box()))
	return;
    out << "% Mark\n";
    strokecol(out, _linecol);
    out<<"% xy " << to_ipe_coord(vt) << "\n";
    float viewing_size = _viewing_xsize>_viewing_ysize ?
	    _viewing_xsize : _viewing_ysize;
    float radius = _vertexsize * viewing_size/1000.0;
    out<<"% sz "<<to_ipe_size(radius)<<'\n';
    strokecol(out, _linecol);
    switch(_vertexshape) {
    case 0:	// dot
	out<<"% ty 2\n";
	break;
    case 3:	// box
	out<<"% ty 3\n";
	break;
    case 1:	// cross
    default:
	out<<"% ty 5\n";
	break;
    case 2:	// square
	out<<"% ty 4\n";
	break;
    }
    out<<"% End\n";
}

void pl_ipefile::draw(const pl_edge_ifc &e)
{
    page_init();
    if (!do_overlap(e.bbox(), viewing_box()))
	return;
    out << "% Line\n";
    strokecol(out, _linecol);
    out<<"% # 2\n";
    out<<to_ipe_coord(e.begin_vt())<<"\n";
    out<<to_ipe_coord(e.end_vt())<<"\n";
    out<<"% End\n";
}

void pl_ipefile::draw(const pl_pgn_ifc &pgn)
{
    page_init();
    if (!do_overlap(pgn.bbox(), viewing_box()))
	return;
    int i;
    out << "% Line\n";
    if (_areafill== FILL || _areafill== FILLOUT) {
	fillcol(out, _areacol);
    }
    if (_areafill== OUTLINE || _areafill== FILLOUT) {
	strokecol(out, _linecol);
    }
    out<<"% # " << pgn.size() << '\n';
    for (i=0; i<pgn.size(); i++) {
	out<<to_ipe_coord(pgn.vertex(i))<<'\n';
    }
    out<<"% cl\n";
    out<<"% End\n";
}

void pl_ipefile::draw(const pl_disc &disc)
{
    page_init();
    if (!do_overlap(disc, viewing_box()))
	return;
    out << "% Circle\n";
    if (_areafill== FILL || _areafill== FILLOUT) {
	fillcol(out, _areacol);
    }
    if (_areafill== OUTLINE || _areafill== FILLOUT) {
	strokecol(out, _linecol);
    }
    out<<"% xy " << to_ipe_coord(disc.center()) << '\n';
    out<<"% r " << to_ipe_size(disc.radius()) << '\n';
    out<<"% End\n";
}

void pl_ipefile::draw(const pl_arc &arc)
{
    page_init();
    if (!do_overlap(arc.bbox(), viewing_box()))
	return;
    const s_coord to_deg = 180/M_PI;
    page_init();
    out << "% Arc\n";
	strokecol(out, _linecol);
    out<<"% xy " << to_ipe_coord(arc.center()) << '\n';
    out<<"% r " << to_ipe_size(arc.radius()) << '\n';
    out	<<"% ang " << to_deg*arc.small_angle().value() <<' '
	<<to_deg*arc.big_angle().value() << '\n';
    out<<"% End\n";
}

void pl_ipefile::draw_text(const pl_fixedvec &vt, const char *str)
{
    page_init();
    if (!lies_inside(vt, viewing_box()))
	return;
    out<<"% Text\n";
    out<<"% s "<< str<< '\n';
    out<<"% xy " << to_ipe_coord(vt) << '\n';
    out<<"% f 0 10\n";
    strokecol(out, _textcol);
    out<<"% End\n";
}

pl_ipefile::pl_ipefile(ostream &os) : out(os)
{
    init_values();
    ipe_init();
}


pl_ipefile::~pl_ipefile()
{
    out<<flush;
}

void pl_ipefile::clear()
{
    page_init();
    pl_vec min, max;
    min = to_ipe_coord(pl_vec(_viewing_x, _viewing_y));
    max = to_ipe_coord(pl_vec(_viewing_x+_viewing_xsize,
				_viewing_y+_viewing_ysize));
    out << "% Line\n";
	fillcol(out, _backcol);
    out<<"% # 4\n";
    out<< min.x() << ' ' << min.y() << '\n';
    out<< max.x() << ' ' << min.y() << '\n';
    out<< max.x() << ' ' << max.y() << '\n';
    out<< min.x() << ' ' << max.y() << '\n';
    out<<"% cl\n";
    out<<"% End\n";
}

void pl_ipefile::pageready()
{
    page_init();
    out<<"% End\n"<<flush;
    _newpage = 1;
}

void pl_ipefile::ipe_init()
{

    out<<flush;
    out<<"%\\IPEid\n";
    out<<"% Group\n";
    _newpage = 1;
}

void pl_ipefile::page_init()
{
    if (! _newpage)
	return;
    _newpage = 0;
}
