#include <GL/gl.h>
#include <GL/glu.h>
#include <PLAGEO/ogl_smart_renderer.h>

/*
 
This renderer class has the following protected members:
    fitting_modes   _fitting_mode;
    int		    _winx, _winy, _windx, _windy;
    GLint	    _vp_dx, _vp_dy;
    GLdouble	    _width, _height;
    reray<GLuint>   _draw_list;
    bool	    _cleanpage;
    void	    init_page();
    void	    init_projection();

Here we describe their purpose.

First there are those parameters having to do with the size of the window
and of the rendered area. 
There are two areas to take into account: the area (in geometric space) that
is to be rendered and the window that is available for rendering. When the
ratio of height and width of those two areas is not the same, the
_fitting_mode decides what strategy is followed. The following possibilities
exist:
STRETCH: the scene is deformed in the x or the y direction, such that it fills
the window. Geometric segments that are of equal length are rendered with
different lengths, depending on their orientation.
CUT_OFF: Not the whole requested area is rendered. A part at the right or at
the top is cut off.
ADD_EXTRA: More than the requested area is rendered. A part at the top or at
the right is added.
ADD_EMPTY: The requested area is rendered. A part of the window at the top or
at the right is not used for drawing.

This scheme has the consequence that there can be a difference between
requested sizes and actual sizes.
The window - which has top-left coordinates <_winx,_winy> - has width (height)
_windx (_windy). The part that is actually used to render has width (height)
_vp_dx (_vp_dy). `vp' stands for viewport.
The requested viewing area is stored in the base class parameters
_viewing_xsize and _viewing_ysize.
The actual parameters are _width and _height.

_draw_list is a list of OpenGL display list identifiers. Everything is
rendered via a display list. This is used to rerender the scene on request,
for instance in response to an expose event.

_cleanpage is a boolean variable that holds track whether a new page should be
used when the next object is drawn. This variable is set by pageready.

init_page() is called by all routines that draw something. It checks whether
_cleanpage is set, in which case first the window is cleared.

init_projection() computes the actual widths and heights out of the requested
widths and heights.

*/

static void error()
{
    abort();
}

static void use_color(int col)
{
    GLubyte color_comp[3];
    int red, green, blue;
    pl_color_components(red, green, blue, col);
    color_comp[0] = red;
    color_comp[1] = green;
    color_comp[2] = blue;
    glColor3ubv(color_comp);
}


pl_ogl_smart_renderer::pl_ogl_smart_renderer()
{
    init_values();
    _cleanpage = true;
    set_viewport(0, 0, 10, 10);
    _fitting_mode = ADD_EXTRA;
}

pl_ogl_smart_renderer::~pl_ogl_smart_renderer()
{
    pageready();	// make sure that viewport is popped.
}

void pl_ogl_smart_renderer::rerender() const
{
    if (_draw_list.size() == 0)
	return;
    ((pl_ogl_smart_renderer*)this)->init_projection();
    for (int i=0; i< _draw_list.size(); i++) {
	glCallList(_draw_list[i]);
    }
//    if (_cleanpage)
	glFlush();
}

void pl_ogl_smart_renderer::set_viewport(int x, int y, int dx, int dy)
{
    if (dx == 0) dx=1;
    if (dx>0) {
	_winx = x; _windx = dx;
    } else {
	_winx = x+dx; _windx = -dx;
    }
    if (dy == 0) dy=1;
    if (dy>0) {
	_winy = y; _windy = dy;
    } else {
	_winy = y+dy; _windy = -dy;
    }
    _viewing_changed = 1;
}

void pl_ogl_smart_renderer::get_viewport(int &x, int &y, int &dx, int &dy)
{
    x = _winx;
    y = _winy;
    dx = _windx;
    dy = _windy;
}

pl_ogl_smart_renderer::fitting_modes
pl_ogl_smart_renderer::set_fitting_mode(fitting_modes fm)
{
    fitting_modes old = _fitting_mode;
    _fitting_mode = fm;
    _viewing_changed = 1;
    return old;
}

void pl_ogl_smart_renderer::init_projection()
{
    if (_viewing_changed) {
	_viewing_changed = 0;
	_vp_dx = _windx;
	_vp_dy = _windy;
	_width = _viewing_xsize;
	_height = _viewing_ysize;
	if (_vp_dx * _height > _vp_dy * _width) {
	    switch (_fitting_mode) {
	    case STRETCH:
		break;
	    case CUT_OFF:
		_height = (_vp_dy*_width)/_vp_dx;
		break;
	    case ADD_EXTRA:
		_width = (_vp_dx*_height)/_vp_dy;
		break;
	    case ADD_EMPTY:
		_vp_dx = GLint((_vp_dy * _width)/_height);
		break;
	    }
	} else {
	    switch (_fitting_mode) {
	    case STRETCH:
		break;
	    case CUT_OFF:
		_width = (_vp_dx*_height)/_vp_dy;
		break;
	    case ADD_EXTRA:
		_height = (_vp_dy*_width)/_vp_dx;
		break;
	    case ADD_EMPTY:
		_vp_dy = GLint((_vp_dx * _height)/_width);
		break;
	    }
	}
	glViewport(_winx, _winy, _vp_dx, _vp_dy);
	if (_fitting_mode == ADD_EMPTY) {
	    glScissor(_winx, _winy, _vp_dx, _vp_dy);
	    glEnable(GL_SCISSOR_TEST);
	} else {
	    glDisable(GL_SCISSOR_TEST);
	}
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(_viewing_x, _viewing_x+_width,
		    _viewing_y, _viewing_y+_height, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    }
}


bool pl_ogl_smart_renderer::to_scene_coords(pl_vec &result, int wx, int wy)
{
// Converts window coordinates to geometric coordinates.
// Note that X window y-coordinates increase from top to bottom, while OpenGL
// y-coordinates increase from bottom to top.
    init_projection();
    float x, y;
    x = (wx - _winx) * _width / _vp_dx + _viewing_x;
    y = _viewing_y + (_winy + _windy - wy) * _height / _vp_dy;
    result = pl_vec(x, y);
    if (x < _viewing_x || x >= _viewing_x + _viewing_xsize ||
	y < _viewing_y || y >= _viewing_y + _viewing_ysize)
	return false;
    return true;
}

void pl_ogl_smart_renderer::init_page()
{
    init_projection();
    if (!_cleanpage)
	return;
    _cleanpage = false;
    clear();  // is this OK? May be redefined. ????
}


void pl_ogl_smart_renderer::pageready()
{
    if (_cleanpage)
	return;
    _cleanpage = true;
    glFlush();
}

// dirty procedure to quickly pass a vertex to the the GL.
// makes use of the internal representation of a vertex.
static inline void vv(const pl_fixedvec &vt)
{
glVertex2fv(& (s_coord &)(vt.x()));
}

static void pl_draw(const pl_fixedvec &vt, int how, s_coord radius)
{
    pl_vec vtx;
    switch(how) {
    case 0: // dot
	glBegin(GL_POINTS);
	vv(vt);
	glEnd();
	break;
    case 1: // cross
    default:
	vtx.set_x(vt.x());
	vtx.set_y(vt.y() - radius);
	glBegin(GL_LINES);
	vv(vtx);
	vtx.set_y(vt.y() + radius);
	vv(vtx);
	vtx.set_x(vt.x() - radius);
	vtx.set_y(vt.y());
	vv(vtx);
	vtx.set_x(vt.x() + radius);
	vv(vtx);
	glEnd();
	break;
    case 2: // square
	vtx.set_x(vt.x() - radius);
	vtx.set_y(vt.y() - radius);
	glBegin(GL_LINE_LOOP);
	vv(vtx);
	vtx.set_x(vt.x() + radius);
	vv(vtx);
	vtx.set_y(vt.y() + radius);
	vv(vtx);
	vtx.set_x(vt.x() - radius);
	vv(vtx);
	glEnd();
	break;
    case 3: // box
	vtx.set_x(vt.x() - radius);
	vtx.set_y(vt.y() - radius);
	glBegin(GL_QUADS);
	vv(vtx);
	vtx.set_x(vt.x() + radius);
	vv(vtx);
	vtx.set_y(vt.y() + radius);
	vv(vtx);
	vtx.set_x(vt.x() - radius);
	vv(vtx);
	glEnd();
	break;
    }
}

void pl_ogl_smart_renderer::clear()
{
    init_projection();
    for (int i=0; i< _draw_list.size(); i++)
	glDeleteLists(_draw_list[i], 1);
    _draw_list.newsize(0);
    GLuint list_index=glGenLists(1);
    if(list_index!=0) {
	glNewList(list_index, GL_COMPILE_AND_EXECUTE);
    }
    int red, green, blue;
    pl_color_components(red, green, blue, _backcol);
    glClearColor(red/255.0, green/255.0, blue/255.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT);
    if(list_index!=0) {
	glEndList();
	_draw_list.append(list_index);
    }
}

void pl_ogl_smart_renderer::draw(const pl_fixedvec &vt)
{
    init_page();
    GLuint list_index=glGenLists(1);
    if(list_index!=0) {
	glNewList(list_index, GL_COMPILE_AND_EXECUTE);
    }
    use_color(_linecol);
    pl_draw(vt, _vertexshape, _vertexsize);
    if(list_index!=0) {
	glEndList();
	_draw_list.append(list_index);
    }
}

void pl_ogl_smart_renderer::draw(const pl_edge_ifc &e)
{
    init_page();
    GLuint list_index=glGenLists(1);
    if(list_index!=0) {
	glNewList(list_index, GL_COMPILE_AND_EXECUTE);
    }
    use_color(_linecol);
    glBegin(GL_LINES);
    vv(e.begin_vt());
    vv(e.end_vt());
    glEnd();
    if(list_index!=0) {
	glEndList();
	_draw_list.append(list_index);
    }
}

// Routines to draw a polygon.

static GLUtriangulatorObj *triangulator_obj = 0;

struct tess_vtx_info {
    int		index;
//    GLboolean	original;
};

struct tess_state {
    reray<tess_vtx_info>    vertices;
//    GLboolean		    cur_original;
    reray<GLenum>	    errors;
};

static struct tess_state *cur_tess_state;

static void tess_begin(GLenum )
{
    cur_tess_state->vertices.newsize(0);
//    cur_tess_state->cur_original = GL_TRUE;
    cur_tess_state->errors.newsize(0);
}

static void tess_edgeflag(GLboolean )
{
//    cur_tess_state->cur_original = original;
}

static void tess_vertex(void *data)
{
    int * indexp = (int*) data;
    tess_vtx_info one_vertex;
    one_vertex.index = *indexp;
//    one_vertex.original = cur_tess_state->cur_original;
    cur_tess_state->vertices.append(one_vertex);
}

static void tess_error(GLenum errno)
{
    cur_tess_state->errors.append(errno);
}

static void tesselate_polygon(const pl_pgn_ifc &pgn)
{
    static reray<int> indices(0);
    int i, sz;
    if (triangulator_obj == 0) {
	triangulator_obj  = gluNewTess();
	if (triangulator_obj == 0)
	    error();
	gluTessCallback(triangulator_obj, GLU_BEGIN, (void (*)()) tess_begin);
	gluTessCallback(triangulator_obj, GLU_EDGE_FLAG, (void (*)()) tess_edgeflag);
	gluTessCallback(triangulator_obj, GLU_VERTEX, (void (*)()) tess_vertex);
//	gluTessCallback(triangulator_obj, GLU_END, (void (*)()) tess_end);
	gluTessCallback(triangulator_obj, GLU_ERROR, (void (*)()) tess_error);
    }
    sz = pgn.size();
    if (indices.size() < sz) {
	i = indices.size();
	indices.resize(sz);
	for ( ; i<indices.size(); i++)
	    indices[i] = i;
    }
    gluBeginPolygon(triangulator_obj);
    GLdouble vtxdata[3];
    vtxdata[2] = 0.0;
    for (i=0; i<pgn.size(); i++) {
	const pl_fixedvec &vt(pgn.vertex(i));
	vtxdata[0] = vt.x();
	vtxdata[1] = vt.y();
	gluTessVertex(triangulator_obj, vtxdata, &(indices[i]));
    }
    gluEndPolygon(triangulator_obj);
}

static int new_in_cache(const pl_pgn_ifc &pgn)
{
    static reray<pl_vec> vertices;
    int same = 1;
    int i;
    if (pgn.size() == vertices.size()) {
	for (i=0; i< pgn.size(); i++) {
	    if ((pgn.vertex(i).x() != vertices[i].x())) {
		same = 0;
		break;
	    }
	    if ((pgn.vertex(i).y() != vertices[i].y())) {
		same = 0;
		break;
	    }
	}
	    
    } else {
	same = 0;
    }
    if (!same) {
	vertices.newsize(pgn.size());
	for (i=0; i< pgn.size(); ++i) {
	    vertices[i] = pgn.vertex(i);
	}
    }
    return !same;
}

static void pl_draw(const pl_pgn_ifc &pgn, int how)
{
    static tess_state tstate;
    int i;
    switch (how) {
    case 1:
    default:
	{
	if (pgn.size() < 3)
	    break;
	cur_tess_state = &tstate;
	if (new_in_cache(pgn)) {
	    tesselate_polygon(pgn);
	}
	if (tstate.errors.size() != 0)
	    error();
	glBegin(GL_TRIANGLES);
	for(i=0; i<tstate.vertices.size(); i++) {
		vv(pgn.vertex(tstate.vertices[i].index));
	}
	glEnd();
	break;
	}
    case 0:
	glBegin(GL_LINE_LOOP);
	for(i=0; i<pgn.size(); i++) {
		vv(pgn.vertex(i));
	}
	glEnd();
	break;
    }
}


void pl_ogl_smart_renderer::draw(const pl_pgn_ifc &pgn)
{
    init_page();
    GLuint list_index=glGenLists(1);
    if(list_index!=0) {
	glNewList(list_index, GL_COMPILE_AND_EXECUTE);
    }
    if (_areafill == FILL || _areafill == FILLOUT) {
	use_color(_areacol);
	pl_draw(pgn, 1);
    }
    if (_areafill == OUTLINE || _areafill == FILLOUT) {
	use_color(_linecol);
	pl_draw(pgn, 0);
    }
    if(list_index!=0) {
	glEndList();
	_draw_list.append(list_index);
    }
}

void pl_ogl_smart_renderer::draw(const pl_disc &disc)
{
    init_page();
    GLuint list_index=glGenLists(1);
    if(list_index!=0) {
	glNewList(list_index, GL_COMPILE_AND_EXECUTE);
    }
    GLUquadricObj *qobj = gluNewQuadric();
    glPushMatrix();
    gluQuadricNormals(qobj, GLU_NONE);
    glTranslatef(disc.center().x(),  disc.center().y(), 0.);
    switch (_areafill) {
    case FILL:
    case FILLOUT:
    default:
	use_color(_areacol);
	gluDisk( qobj, 0.,  disc.radius(), 64, 1);
	gluDeleteQuadric(qobj);
	if (_areafill == FILL)
	    break;
	// fallthrough otherwise.
    case OUTLINE:
	use_color(_linecol);
	gluQuadricDrawStyle(qobj, GLU_SILHOUETTE);
	gluDisk( qobj, 0.,  disc.radius(), 64, 1);
        break;
    }
    glPopMatrix();
    gluDeleteQuadric(qobj);
    if(list_index!=0) {
	glEndList();
	_draw_list.append(list_index);
    }
}

void pl_ogl_smart_renderer::draw(const pl_arc &arc)
{
    GLuint list_index=glGenLists(1);
    if(list_index!=0) {
	glNewList(list_index, GL_COMPILE_AND_EXECUTE);
    }
    static reray<pl_vec> approximation;
    int i;
    approximation.newsize(0);
    arc.inner_approximation_n(approximation, 64);
    glBegin(GL_LINE_STRIP);
    for (i=0; i<approximation.size(); i++)
	vv(approximation[i]);
    glEnd();
    if(list_index!=0) {
	glEndList();
	_draw_list.append(list_index);
    }
/*
    const s_coord to_deg = 180/M_PI;
    GLUquadricObj *qobj = gluNewQuadric();
    gluQuadricNormals(qobj, GLU_NONE);
    gluQuadricDrawStyle(qobj, GLU_SILHOUETTE);
    glPushMatrix();
    glTranslatef(arc.center().x(),  arc.center().y(),  0.);
    gluPartialDisk( qobj, 0.,  arc.radius(), 32, 1,
	 to_deg*arc.small_angle().normalised_value(0),
	   to_deg*(arc.big_angle().normalised_value(0)-
	   arc.small_angle().normalised_value(0)));
    glPopMatrix();
    gluDeleteQuadric(qobj);
*/
}

void pl_ogl_smart_renderer::draw_text(const pl_fixedvec&, const char*)
{
    
}
