#include <stdlib.h>
#include <assert.h>
#include <PLAGEO/sweeplist.h>


struct pl_sweeplist_node {
		    pl_sweeplist_node(int l);
		    ~pl_sweeplist_node();
    int		    level;
    void *	    info;
    pl_sweeplist_node  **forward,
		    **backward;
private:
    void	operator=(const pl_sweeplist_node &);
};


pl_sweeplist_node::pl_sweeplist_node(int l)
{
    level = l;
    forward = new pl_sweep_item[l];
    backward = new pl_sweep_item[l];
}

pl_sweeplist_node::~pl_sweeplist_node()
{
    delete [] forward;
    delete [] backward;
}


pl_sweeplist::pl_sweeplist(int (*search_func)(const pl_vec &, void *))
{
    _highest = 0;
    _cmp = search_func;
}

bool pl_sweeplist::is_empty() const
{
//  assert(true);
    return _highest == 0;
}

int pl_sweeplist::search(pl_sweep_item &result, const pl_vec & key) const
{
    assert(!is_empty());
    assert(is_valid_search_key(key));

    switch(_cmp(key, _highest->info)) {
    case -1:
	return search_before(result, key, _highest, _highest->level-1);
    case 0:
	result = _highest;
	return 0;
    case 1:
	return search_after(result, key, _highest, _highest->level-1);
    default: // should not happen
	assert(false);
	return -1;
    }
}

int pl_sweeplist::search_before(pl_sweep_item &result, const pl_vec & key, 
		    pl_sweep_item cur1, int level) const
{
    if (level == -1) {
	result = cur1;
	return -1;
    }
    pl_sweep_item cur = cur1;
    pl_sweep_item next;
    next = cur->backward[level];
    if (next == 0) {
	return search_before(result, key, cur, level-1);
    } else {
	switch(_cmp(key, next->info)) {
	case -1:
	    return search_before(result, key, next, level);
	case 0:
	    result = next;
	    return 0;
	case 1:
	    return search_after(result, key, next, level-1);
	default: // should not happen
	    assert(false);
	    return -1;
	}
    }
}

int pl_sweeplist::search_after(pl_sweep_item &result, const pl_vec & key, 
		    pl_sweep_item cur1, int level) const
{
    if (level == -1) {
	result = cur1;
	return 1;
    }
    pl_sweep_item cur = cur1;
    pl_sweep_item next;
    next = cur->forward[level];
    if (next == 0) {
	return search_after(result, key, cur, level-1);
    } else {
	switch(_cmp(key, next->info)) {
	case -1:
	    return search_before(result, key, next, level-1);
	case 0:
	    result = next;
	    return 0;
	case 1:
	    return search_after(result, key, next, level);
	default: // should not happen
	    assert(false);
	    return -1;
	}
    }
}

pl_sweep_item pl_sweeplist::first() const
{
    assert(!is_empty());
    pl_sweep_item cur, next;
    int i;
    cur = _highest;
    i = _highest->level-1;
    while (i>=0) {
	next = cur->backward[i];
	if (next == 0)
	    i--;
	else
	    cur = next;
    }
    return cur;
}

pl_sweep_item pl_sweeplist::last() const
{
    assert(!is_empty());
    pl_sweep_item cur, next;
    int i;
    cur = _highest;
    i = _highest->level-1;
    while (i >= 0) {
	next = cur->forward[i];
	if (next == 0)
	    i--;
	else
	    cur = next;
    }
    return cur;
}

pl_sweep_item pl_sweeplist::pred(pl_sweep_item it) const
{
    assert(is_valid_item(it));
    return it->backward[0];
}

pl_sweep_item pl_sweeplist::succ(pl_sweep_item it) const
{
    assert(is_valid_item(it));
    return it->forward[0];
}

void *pl_sweeplist::info(pl_sweep_item it) const
{
    assert(is_valid_item (it));
    return it->info;
}

bool pl_sweeplist::is_valid_item(pl_sweep_item it) const
{
//    assert(true);
    pl_sweep_item cur;
    cur = _highest;
    while (cur != 0) {
	if (cur == it)
	    return true;
	cur = cur->forward[0];
    }
    cur = _highest;
    while (cur != 0) {
	if (cur == it)
	    return true;
	cur = cur->backward[0];
    }
    return false;
}

bool pl_sweeplist::is_valid_search_key(const pl_vec & key) const
{
    if (is_empty())
	return false;
    pl_sweep_item cur;
    int lastcmp, cmp;
    cur = first();
    lastcmp = _cmp(key, info(cur));
    assert(lastcmp == -1 || lastcmp == 0 || lastcmp == 1);
    while (1) {
	cur = cur->forward[0];
	if (cur == 0)
	    break;
	cmp =  _cmp(key, info(cur));
	assert(cmp == -1 || cmp == 0 || cmp == 1);
	if (lastcmp < cmp)
	    return false;
    }
    return true;
}

pl_sweep_item pl_sweeplist::insert_first(void *info)
{
    assert(is_empty());
    _highest = new pl_sweeplist_node(1);
    _highest->forward[0] = 0;
    _highest->backward[0] = 0;
    _highest->info = info;
    assert(is_consistent());
    return _highest;
}

static bool next_level()
{
    const double CUT = 0.5;
    static unsigned short xsubi[3];
    static bool initialised = false;
    if (!initialised) {
	initialised = true;
        xsubi[0] = 0x8f4a;
    }
    return ( erand48(xsubi) < CUT);
}

pl_sweep_item pl_sweeplist::new_pl_sweep_item()
{
    assert(!is_empty());
    int newlevel;
    // choose a depth with a random proces;
    newlevel = 1;
    while (newlevel <= _highest->level && next_level())
	newlevel++;
    return new pl_sweeplist_node(newlevel);
}

pl_sweep_item pl_sweeplist::insert_before(pl_sweep_item after, void *info)
{
    assert(is_valid_item (after));
    pl_sweep_item result, cur;
    int startlevel, i;
    result = new_pl_sweep_item();
    result->info = info;
// now fill in level pointers
    startlevel = 0;
    i = startlevel;
    cur = after;
// invariant: all backward pointers of result in [0, startlevel) are correct.
//	      all forward pointers of result in [0, i) are correct.
    while (i < result->level) {
	if (cur == 0) {
	    result->forward[i] = 0;
	    i++;
	} else {
	    if (cur->level <= i) {
		cur = cur->forward[i-1];
	    } else {
		result->forward[i] = cur;
		result->backward[i] = cur->backward[i];
		if (cur->backward[i] != 0) {
		    cur->backward[i]->forward[i] = result;
		}
		cur->backward[i] = result;
		i++;
		startlevel = i;
	    }
	}
    }
    i = startlevel;
    cur = result->backward[i-1];
// invariant: all backward pointers of result in [0, i) are correct
    while (i < result->level) {
	if (cur == 0) {
	    result->backward[i] = 0;
	    i++;
	} else {
	    if (cur->level <= i) {
		cur = cur->backward[i-1];
	    } else {
		result->backward[i] = cur;
		cur->forward[i] = result;
		i++;
	    }
	}
    }
// check whether newly inserted node is highest node.
    if (result->level > _highest->level
	|| (result->level == _highest->level
		    && result->backward[result->level-1]==0))
	_highest = result;
    assert(is_consistent());
    return result;
}

pl_sweep_item pl_sweeplist::insert_after(pl_sweep_item before, void *info)
{
    assert(is_valid_item(before));
    pl_sweep_item result, cur;
    int startlevel, i;
    result = new_pl_sweep_item();
    result->info = info;
// now fill in level pointers
    startlevel = 0;
    i = startlevel;
    cur = before;
// invariant: all forward pointers of result in [0, startlevel) are correct.
//	      all backward pointers of result in [0, i) are correct.
    while (i < result->level) {
	if (cur == 0) {
	    result->backward[i] = 0;
	    i++;
	} else {
	    if (cur->level <= i) {
		cur = cur->backward[i-1];
	    } else {
		result->backward[i] = cur;
		result->forward[i] = cur->forward[i];
		if (cur->forward[i] != 0) {
		    cur->forward[i]->backward[i] = result;
		}
		cur->forward[i] = result;
		i++;
		startlevel = i;
	    }
	}
    }
    i = startlevel;
    cur = result->forward[i-1];
// invariant: all forward pointers of result in [0, i) are correct.
    while (i < result->level) {
	if (cur == 0) {
	    result->forward[i] = 0;
	    i++;
	} else {
	    if (cur->level <= i) {
		cur = cur->forward[i-1];
	    } else {
		result->forward[i] = cur;
		cur->backward[i] = result;
		i++;
	    }
	}
    }
// check whether newly inserted node is highest node.
    if (result->level > _highest->level
	|| (result->level == _highest->level
		    && result->backward[result->level-1]==0))
	_highest = result;
    assert(is_consistent());
    return result;
}

pl_sweep_item pl_sweeplist::find_and_insert(const pl_vec & key, void *info)
{
    assert(is_empty() || is_valid_search_key(key));

    if (is_empty())
	return insert_first(info);
    else {
	pl_sweep_item cur;
	int cmp = search(cur, key);
	if (cmp<0)
	    return insert_before(cur, info);
	else
	    return insert_after(cur, info);
    }
}

static pl_sweep_item new_highest(pl_sweep_item old)
{
// precondition: list will not be empty.
    pl_sweep_item cur;
    int i = old->level-1;
    cur = old->forward[i];
    while (cur == 0) {
	i--;
	cur = old->backward[i];
	if (cur == 0) {
	    cur = old->forward[i];
	} else {
	    while (cur->backward[i] != 0)
		cur = cur->backward[i];
	}
    }
    return cur;
}

void pl_sweeplist::remove(pl_sweep_item it)
{
    assert(is_valid_item (it));
    int i;
    if (it->forward[0] == 0  && it->backward[0] == 0) {
	_highest = 0;
	delete it;
	return;
    }
    for (i=0; i<it->level; i++) {
	pl_sweep_item before, after;
	before = it->backward[i];
	after = it->forward[i];
	if (before != 0)
	    before->forward[i] = after;
	if (after != 0)
	    after->backward[i] = before;
    }
    if (it == _highest) {
	_highest = new_highest(it);
    }
    delete it;
    assert(is_consistent());
}

static bool is_a_predecessor_of(pl_sweep_item it1, pl_sweep_item it2)
{
    pl_sweep_item newit = it2->backward[0];
    while (newit != 0) {
	if (newit == it1)
	    return true;
	newit = newit->backward[0];
    }
    return false;
}

void pl_sweeplist::exchange(pl_sweep_item it1, pl_sweep_item it2)
{
    assert(is_valid_item(it1) && is_valid_item(it2));
    assert(is_a_predecessor_of(it1, it2));

    pl_sweep_item cur;
    int fill_level = it1->level;
    if (it1->level < it2->level) {
	// adapt the backward and forward pointers of it2 that go over it1.
	// Only if there are nodes higher than it1 between it1 and it2 there
	// is some real work to do.
	fill_level = it1->level;
	cur = it1->forward[fill_level-1];
	while (cur != it2 && fill_level < it2->level) {
	    if (cur->level == fill_level) {
		cur = cur->forward[fill_level-1];
	    } else {	// cur->level > fill_level.
		it2->backward[fill_level]->forward[fill_level]
			= it2->forward[fill_level];
		it2->backward[fill_level] = cur->backward[fill_level];
		cur->backward[fill_level] = it2;
		it2->forward[fill_level] = cur;
		fill_level++;
	    }
	}
	fill_level = it1->level;
    }
    if (it1->level > it2->level) {
	// adapt the backward and forward pointers of it1 that go over it2.
	fill_level = it2->level;
	cur = it2->backward[fill_level-1];
	while (cur != it1 && fill_level < it1->level) {
	    if (cur->level == fill_level) {
		cur = cur->backward[fill_level-1];
	    } else {	// cur->level > fill_level.
		it1->forward[fill_level]->backward[fill_level]
			= it1->backward[fill_level];
		it1->forward[fill_level] = cur->forward[fill_level];
		cur->forward[fill_level] = it1;
		it1->backward[fill_level] = cur;
		fill_level++;
	    }
	}
	fill_level = it2->level;
    }
    // now fill in lower pointers
    pl_sweep_item pred, succ;
    while (fill_level> 0) {
	fill_level--;
	pred = it1->backward[fill_level];
	succ = it2->forward[fill_level];
	if (it1->forward[fill_level] == it2) {
	    it1->backward[fill_level] = it2;
	    it2->forward[fill_level] = it1;
	} else {
	    it1->backward[fill_level] = it2->backward[fill_level];
	    it2->forward[fill_level] = it1->forward[fill_level];
	    it1->backward[fill_level]->forward[fill_level] = it1;
	    it2->forward[fill_level]->backward[fill_level] = it2;
	}
	it1->forward[fill_level] = succ;
	it2->backward[fill_level] = pred;
	if (pred != 0)
	    pred->forward[fill_level] = it2;
	if (succ != 0)
	    succ->backward[fill_level] = it1;
    }
// The exchange may have invalidated the _highest node (which must be the
// first of the nodes with biggest height).
    while (_highest->backward[_highest->level-1] != 0)
	_highest = _highest->backward[_highest->level-1];
    assert(is_consistent());
}

void pl_sweeplist::clear()
{
    if (is_empty())
	return;
    pl_sweep_item cur, next;
    cur = _highest;
    next = cur->backward[0];
    while(next != 0) {
	cur = next;
        next = cur->backward[0];
	delete cur;
    }
    cur = _highest;
    while (cur != 0) {
	next = cur->forward[0];
	delete cur;
	cur = next;
    }
    _highest = 0;
}

pl_sweeplist::~pl_sweeplist()
{
    // delete alle nodes
    clear();
}

bool pl_sweeplist::is_consistent() const
{
    pl_sweep_item cur, other;
    int i;
    if (_highest == 0)
	return true;
    cur = _highest;
    for (i=0; i<cur->level; i++) {
	other = cur->forward[i];
	if (other == 0)
	    continue;
	if (other->backward[i] != cur)
	    return false;
    }
    for (i=0; i<cur->level; i++) {
	other = cur->backward[i];
	if (other == 0)
	    continue;
	if (other->forward[i] != cur)
	    return false;
    }
    for (cur=_highest->forward[0]; cur != 0; cur = cur->forward[0]) {
	for (i=0; i<cur->level; i++) {
	    other = cur->forward[i];
	    if (other == 0)
		continue;
	    if (other->backward[i] != cur)
		return false;
	}
	for (i=0; i<cur->level; i++) {
	    other = cur->backward[i];
	    if (other == 0)
		return false;
	    if (other->forward[i] != cur)
		return false;
	}
    }
    for (cur=_highest->backward[0]; cur != 0; cur = cur->backward[0]) {
	for (i=0; i<cur->level; i++) {
	    other = cur->forward[i];
	    if (other == 0)
		return false;
	    if (other->backward[i] != cur)
		return false;
	}
	for (i=0; i<cur->level; i++) {
	    other = cur->backward[i];
	    if (other == 0)
		continue;
	    if (other->forward[i] != cur)
		return false;
	}
    }
    if (_highest->backward[_highest->level-1] != 0)
	return false;
    return true;
}
